@@ -25,6 +25,7 @@ from secure_delete import secure_delete
 from threading import Thread
 import colemen_string_utils as strUtils
 import logging
+import ftputil
 
 logger = logging.getLogger(__name__)
 
@@ -118,7 +119,7 @@ def _compress_single_file_gzip(file_path):
     return success
 
 
-def exists(file_path):
+def exists(file_path, **kwargs):
     '''
         Confirms that the file exists.
 
@@ -130,6 +131,15 @@ def exists(file_path):
         `return` {bool}
             True if the file exists, False otherwise.
     '''
+
+    ftp = obj.get_kwarg(["ftp"], None, None, **kwargs)
+
+    if ftp is not None:
+        if ftp.path.isFile(file_path):
+            if ftp.path.exists(file_path):
+                return True
+        return False
+
     if os.path.isfile(file_path) is True:
         return True
     else:
@@ -164,25 +174,27 @@ def delete(file_path, **kwargs):
         `version`: 1.0
         `method_name`: delete
     '''
+    ftp = obj.get_kwarg(["ftp"], None, None, **kwargs)
     shred = obj.get_kwarg(["shred", "secure"], False, (bool), **kwargs)
     threading = obj.get_kwarg(["threading"], True, (bool), **kwargs)
     max_threads = obj.get_kwarg(["max_threads"], 15, (int), **kwargs)
     min_thread_threshold = obj.get_kwarg(["min thread threshold", "min files thread"], 100, (int), **kwargs)
 
     file_list = _gen_path_list(file_path)
-    print("file_list: ", json.dumps(file_list, indent=4))
+    # print("file_list: ", json.dumps(file_list, indent=4))
     # exit()
     if len(file_list) > 0:
+
         if threading is True and len(file_list) >= min_thread_threshold:
-            _thread_file_action(file_list, action="delete", max_threads=max_threads, min_thread_threshold=min_thread_threshold, shred=shred)
+            _thread_file_action(file_list, action="delete", max_threads=max_threads, min_thread_threshold=min_thread_threshold, shred=shred, ftp=ftp)
             return
 
         for file in file_list:
             if isinstance(file, (dict)):
                 if 'file_path' in file:
-                    _delete_single_file(file['file_path'], shred)
+                    _delete_single_file(file['file_path'], shred, ftp)
             if isinstance(file, (str)):
-                _delete_single_file(file, shred)
+                _delete_single_file(file, shred, ftp)
     return True
 
 
@@ -384,7 +396,7 @@ def _copy_files_from_array(file_list, attempted_delete=False):
     return True
 
 
-def _delete_single_file(file_path, shred=False):
+def _delete_single_file(file_path, shred=False, ftp=None):
     '''
         Deletes or shreds a single file.
 
@@ -396,6 +408,8 @@ def _delete_single_file(file_path, shred=False):
             Path to the file to delete.
         [`shred`=False] {bool}
             Securely shred the file (slower)
+        [`ftp`=None] {ftputil object}
+            The ftp object to use.
 
         Return {bool}
         ----------------------
@@ -410,6 +424,9 @@ def _delete_single_file(file_path, shred=False):
         `method_name`: _delete_single_file
     '''
     success = False
+    if ftp.path.exists(file_path) is True:
+        ftp.remove(file_path)
+        return
     if exists(file_path) is True:
         try:
             if shred is True:
@@ -428,7 +445,7 @@ def _delete_single_file(file_path, shred=False):
     return success
 
 
-def _delete_files_from_array(file_list, shred=False):
+def _delete_files_from_array(file_list, shred=False, ftp=None):
     '''
         Delete a list of files.
 
@@ -456,8 +473,8 @@ def _delete_files_from_array(file_list, shred=False):
     '''
     success = True
     for file in file_list:
-        if exists(file):
-            result = _delete_single_file(file, shred=shred)
+        if exists(file, ftp=ftp):
+            result = _delete_single_file(file, shred=shred, ftp=None)
             if result is not True:
                 success = False
     return success
@@ -468,6 +485,7 @@ def _thread_file_action(file_list, **kwargs):
     min_thread_threshold = obj.get_kwarg(["min thread threshold", "min files thread"], 100, (int), **kwargs)
     action = obj.get_kwarg(["action"], "copy", (str), **kwargs)
     shred = obj.get_kwarg(["shred"], False, (bool), **kwargs)
+    ftp = obj.get_kwarg(["ftp"], None, None, **kwargs)
 
     if len(file_list) <= min_thread_threshold:
         max_threads = 1
@@ -483,7 +501,7 @@ def _thread_file_action(file_list, **kwargs):
         if action == "copy":
             threads.append(Thread(target=_copy_files_from_array, args=(files_array,)))
         if action == "delete":
-            threads.append(Thread(target=_delete_files_from_array, args=(files_array, shred)))
+            threads.append(Thread(target=_delete_files_from_array, args=(files_array, shred, ftp)))
 
     for thread in threads:
         thread.start()
@@ -547,6 +565,20 @@ def get_data(file_path, **kwargs):
         `method_name`: get_data
     '''
 
+    ftp = obj.get_kwarg(['ftp'], None, None, **kwargs)
+    if ftp is not None:
+        file_data = {}
+        path = strUtils.format.file_path(file_path, url=True)
+        lstat = ftp.path.lstat(path)
+        file_data['file_path'] = path
+        file_data['file_name'] = ftp.path.basename(path)
+        file_data['extension'] = get_ext(path)
+        file_data['name_no_ext'] = get_name_no_ext(file_path)
+        file_data['dir_path'] = ftp.path.dirname(file_path)
+        file_data['modified_time'] = datetime.timestamp(datetime.utcfromtimestamp(lstat[8]))
+        file_data['size'] = lstat[6]
+        return file_data
+
     data_include = obj.get_kwarg(['include', "data include"], [], (list, str), **kwargs)
     if isinstance(data_include, (str)):
         data_include = [data_include]
@@ -585,7 +617,7 @@ def get_data(file_path, **kwargs):
         logger.warning(f"Failed to find the file: {file_path}")
 
 
-def get_modified_time(file_path):
+def get_modified_time(file_path, ftp=None):
     '''
         get the modified from the file
 
@@ -608,8 +640,12 @@ def get_modified_time(file_path):
         `version`: 1.0
         `method_name`: get_modified_time
     '''
+    if ftp is not None:
+        return datetime.timestamp(datetime.utcfromtimestamp(ftp.path.getmtime(file_path)))
+
     mod_time = os.path.getmtime(file_path)
-    return int(datetime.fromtimestamp(mod_time).replace(tzinfo=timezone.utc).timestamp())
+    mod_time = int(datetime.fromtimestamp(mod_time).replace(tzinfo=timezone.utc).timestamp())
+    return datetime.timestamp(datetime.fromtimestamp(mod_time))
 
 
 def get_access_time(file_path):
@@ -723,6 +759,162 @@ def get_name_no_ext(file_path):
     return os.path.basename(file_path).replace(get_ext(file_path), '')
 
 
+def gen_relative_path(src_base, dst_base, file_path):
+    '''
+        Finds the relative path from the src_base, dst_base and file path.
+
+        example: 
+            file_path = r"Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php\\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php"\n
+            src_base = r"Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php"\n
+            dst_base = r"/equari_php/\n
+            relative_path = \\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php
+
+        ----------
+
+        Arguments
+        -------------------------
+        `src_base` {str}
+            The path to the root directory of the "source".
+        `dst_base` {str}
+            The path to the root directory of the "destination".
+        `file_path` {str}
+            The path of the file to generate a relative path for.
+
+        Return {string|None}
+        ----------------------
+        The relative path or None if one cannot be found.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 12-17-2021 16:14:55
+        `memberOf`: file
+        `version`: 1.0
+        `method_name`: gen_relative_path
+    '''
+    src_base = strUtils.format.file_path(src_base)
+    dst_base = strUtils.format.file_path(dst_base)
+    file_path = strUtils.format.file_path(file_path)
+
+    src_len = len(src_base)
+    dst_len = len(dst_base)
+    relative_path = None
+
+    if src_base == dst_base:
+        print("src_base is identical to dst_base")
+        return relative_path
+    if src_base not in file_path and dst_base not in file_path:
+        print("The file path must originate from either the src_path or dst_path.")
+        return relative_path
+
+    if src_len > dst_len:
+        dst_pos = file_path.lower().find(dst_base.lower())
+        if dst_pos != -1:
+            # print(f"file originates from dst_base: {dst_pos}")
+            if dst_pos == 0:
+                relative_path = file_path[len(dst_base):]
+            else:
+                relative_path = file_path[dst_pos:]
+
+            # print(f"relative_path: {file_path[dst_pos:]}")
+
+    if dst_len > src_len:
+        src_pos = file_path.lower().find(src_base.lower())
+        if src_pos != -1:
+            # print(f"file originates from src_base: {src_pos}")
+            if src_pos == 0:
+                relative_path = file_path[len(src_base):]
+            else:
+                relative_path = file_path[src_pos:]
+
+    relative_path = relative_path.replace(dst_base, "")
+    relative_path = relative_path.replace(src_base, "")
+    return relative_path
+
+
+def gen_dst_path(src_base, dst_base, file_path):
+    '''
+        Generates the destination path for a file.
+
+
+        example: 
+            file_path = r"Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php\\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php"\n
+            src_base = r"Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php"\n
+            dst_base = r"/equari_php/\n
+            returns:\n
+            \\equari_php\\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php
+
+        ----------
+
+        Arguments
+        -------------------------
+        `src_base` {str}
+            The path to the root directory of the "source".
+        `dst_base` {str}
+            The path to the root directory of the "destination".
+        `file_path` {str}
+            The path of the file to generate a relative path for.
+
+        Return {string|None}
+        ----------------------
+        The destination path to the file or None if one cannot be found.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 12-17-2021 16:21:22
+        `memberOf`: file
+        `version`: 1.0
+        `method_name`: gen_dst_path
+    '''
+
+    if src_base not in file_path and dst_base not in file_path:
+        return strUtils.format.file_path(f"{dst_base}/{file_path}")
+    relative_path = gen_relative_path(src_base, dst_base, file_path)
+    return strUtils.format.file_path(f"{dst_base}/{relative_path}")
+
+
+def gen_src_path(src_base, dst_base, file_path):
+    '''
+        Generates the source path for a file.
+
+
+        example: 
+            src_base = r"Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php"\n
+            dst_base = r"/equari_php/"\n
+            file_path = r"\\equari_php\\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php"\n
+            returns:\n
+            Z:\\Structure\\Ra9\\2021\\21-0134_EquariServer\\equari_php\\equari_php\\pcom\\animal\\read\\PCOM_getOMNIAnimalData.php
+
+        ----------
+
+        Arguments
+        -------------------------
+        `src_base` {str}
+            The path to the root directory of the "source".
+        `dst_base` {str}
+            The path to the root directory of the "destination".
+        `file_path` {str}
+            The path of the file to generate a relative path for.
+
+        Return {string|None}
+        ----------------------
+        The source path to the file or None if one cannot be found.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 12-17-2021 16:21:22
+        `memberOf`: file
+        `version`: 1.0
+        `method_name`: gen_dst_path
+    '''
+    if src_base not in file_path and dst_base not in file_path:
+        return strUtils.format.file_path(f"{src_base}/{file_path}")
+    relative_path = gen_relative_path(src_base, dst_base, file_path)
+    return strUtils.format.file_path(f"{src_base}/{relative_path}")
+
+
 # file = r"C:\Users\Colemen\Desktop\DAZDOW~1\poses\STANDI~1\IM0008~1\Content\People\GENESI~2\Poses\AEONSO~1\STANDI~1\LIMBSL~1\SC-WE'~2.DUF"
 # file = r"C:\\Users\\Colemen\\Desktop\\DAZ DOWNLOADS\\poses\\Standing Conversation Poses for Genesis 8\\IM00083571-01_StandingConversationPosesforGenesis8\\Content\\People\\Genesis 8 Male\\Poses\\Aeon Soul\\Standing Conversation\\Limbs Legs\\SC-We're all in the same boat Legs-M Genesis 8 Male.duf"
 # clean = clean_path(file)
